@@ -16,7 +16,7 @@ use bitflags::bitflags;
 use libc::{c_char, c_int};
 
 use std::cell::RefCell;
-use std::ffi::{CStr, CString};
+use std::ffi::{c_void, CStr, CString};
 use std::fmt::{self, Display, Formatter};
 use std::marker::{self, PhantomData};
 use std::mem;
@@ -81,6 +81,14 @@ pub trait BaseEvent: Raw<xcb_generic_event_t> {
 
     /// The number associated to this event
     const NUMBER: u32;
+
+    /// Check whether this event was emitted by the SendEvent request
+    /// See `[crate::x::SendEvent]`.
+    fn is_from_send_event(&self) -> bool {
+        let ev = self.as_raw();
+        let response_type = unsafe { (*ev).response_type };
+        (response_type & 0x80) != 0
+    }
 }
 
 /// A trait for GE_GENERIC events
@@ -451,14 +459,14 @@ pub struct AuthInfo<'a> {
 }
 
 /// Display info returned by [`parse_display`]
-#[derive(Debug)]
+#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
 pub struct DisplayInfo {
     /// The hostname
-    host: String,
+    pub host: String,
     /// The display number
-    display: i32,
+    pub display: i32,
     /// The screen number
-    screen: i32,
+    pub screen: i32,
 }
 
 /// Parses a display string in the form documented by [X (7x)](https://linux.die.net/man/7/x).
@@ -469,42 +477,63 @@ pub struct DisplayInfo {
 /// If `name` empty, it uses the environment variable `DISPLAY`.
 ///
 /// If `name` does not contain a screen number, `DisplayInfo::screen` is set to `0`.
+///
+/// # Example
+/// ```
+/// use xcb::{DisplayInfo, parse_display};
+///
+/// assert_eq!(parse_display(":0"), Some(DisplayInfo {
+///     host: "".to_string(),
+///     display: 0,
+///     screen: 0,
+/// }));
+/// assert_eq!(parse_display("localhost:0.1"), Some(DisplayInfo {
+///     host: "localhost".to_string(),
+///     display: 0,
+///     screen: 1,
+/// }));
+///
+/// assert!(parse_display("0").is_none());
+/// ```
 pub fn parse_display(name: &str) -> Option<DisplayInfo> {
-    unsafe {
-        let name = CString::new(name).unwrap();
-        let mut hostp: *mut c_char = ptr::null_mut();
-        let mut display = 0i32;
-        let mut screen = 0i32;
+    let name = CString::new(name).unwrap();
+    let mut hostp: *mut c_char = ptr::null_mut();
+    let mut display = 0i32;
+    let mut screen = 0i32;
 
-        let success = xcb_parse_display(
+    let success = unsafe {
+        xcb_parse_display(
             name.as_ptr(),
             &mut hostp as *mut _,
             &mut display as *mut _,
             &mut screen as *mut _,
-        );
+        )
+    };
 
-        if success != 0 {
-            let host = CStr::from_ptr(hostp as *const _)
-                .to_str()
-                .unwrap()
-                .to_string();
+    if success != 0 {
+        let host = unsafe { CStr::from_ptr(hostp as *const _) }
+            .to_str()
+            .unwrap()
+            .to_string();
 
+        unsafe {
             libc::free(hostp as *mut _);
-
-            Some(DisplayInfo {
-                host,
-                display,
-                screen,
-            })
-        } else {
-            None
         }
+
+        Some(DisplayInfo {
+            host,
+            display,
+            screen,
+        })
+    } else {
+        None
     }
 }
 
 /// A struct that serve as an identifier for internal special queue in XCB
 ///
 /// See [Connection::register_for_special_xge].
+#[deprecated(note = "Broken API: use `SpecialEvent` instead")]
 #[cfg(any(feature = "xinput", feature = "present"))]
 #[derive(Debug)]
 pub struct SpecialEventId {
@@ -512,6 +541,7 @@ pub struct SpecialEventId {
     stamp: Timestamp,
 }
 
+#[allow(deprecated)]
 #[cfg(any(feature = "xinput", feature = "present"))]
 impl SpecialEventId {
     /// The X timestamp associated with this special event Id
@@ -520,6 +550,22 @@ impl SpecialEventId {
     }
 }
 
+/// A struct that serve as an identifier for internal special queue in XCB
+///
+/// See [Connection::register_for_special_event].
+#[cfg(any(feature = "xinput", feature = "present"))]
+#[derive(Debug)]
+pub struct SpecialEvent {
+    raw: *mut xcb_special_event_t,
+}
+
+// safe because XCB is thread safe.
+#[cfg(any(feature = "xinput", feature = "present"))]
+unsafe impl Send for SpecialEvent {}
+
+#[cfg(any(feature = "xinput", feature = "present"))]
+unsafe impl Sync for SpecialEvent {}
+
 /// Error type that is returned by `Connection::has_error`.
 #[derive(Debug)]
 pub enum ConnError {
@@ -1162,17 +1208,6 @@ impl Connection {
         }
     }
 
-    unsafe fn handle_wait_for_event(&self, ev: *mut xcb_generic_event_t) -> Result<Event> {
-        if ev.is_null() {
-            self.has_error()?;
-            panic!("xcb_wait_for_event returned null with I/O error");
-        } else if is_error(ev) {
-            Err(error::resolve_error(ev as *mut _, &self.ext_data).into())
-        } else {
-            Ok(event::resolve_event(ev, &self.ext_data))
-        }
-    }
-
     /// Resolve an xcb_generic_event_t pointer into an Event.
     /// # Safety
     /// The caller is repsonsible to ensure that the `ev` pointer is not NULL.
@@ -1183,17 +1218,6 @@ impl Connection {
         event::resolve_event(ev, &self.ext_data)
     }
 
-    unsafe fn handle_poll_for_event(&self, ev: *mut xcb_generic_event_t) -> Result<Option<Event>> {
-        if ev.is_null() {
-            self.has_error()?;
-            Ok(None)
-        } else if is_error(ev) {
-            Err(error::resolve_error(ev as *mut _, &self.ext_data).into())
-        } else {
-            Ok(Some(event::resolve_event(ev, &self.ext_data)))
-        }
-    }
-
     /// Blocks and returns the next event or error from the server.
     ///
     /// # Example
@@ -1278,7 +1302,9 @@ impl Connection {
     /// XGE events are only defined in the `xinput` and `present` extensions
     ///
     /// This function is present only if either of the `xinput` or `present` cargo features are active.
+    #[deprecated(note = "Broken API: use `register_for_special_event` instead")]
     #[cfg(any(feature = "xinput", feature = "present"))]
+    #[allow(deprecated)]
     pub fn register_for_special_xge<XGE: GeEvent>(&self) -> SpecialEventId {
         unsafe {
             let ext: *mut xcb_extension_t = match XGE::EXTENSION {
@@ -1298,15 +1324,19 @@ impl Connection {
     }
 
     /// Stop listening to a special event
+    #[deprecated(note = "use `unregister_for_special_event` instead")]
     #[cfg(any(feature = "xinput", feature = "present"))]
+    #[allow(deprecated)]
     pub fn unregister_for_special_xge(&self, se: SpecialEventId) {
         unsafe {
-            xcb_unregister_for_special_xge(self.c, se.raw);
+            xcb_unregister_for_special_event(self.c, se.raw);
         }
     }
 
     /// Returns the next event from a special queue, blocking until one arrives
+    #[deprecated(note = "Broken API: use `wait_for_special_event2` instead")]
     #[cfg(any(feature = "xinput", feature = "present"))]
+    #[allow(deprecated)]
     pub fn wait_for_special_event(&self, se: SpecialEventId) -> Result<Event> {
         unsafe {
             let ev = xcb_wait_for_special_event(self.c, se.raw);
@@ -1315,7 +1345,9 @@ impl Connection {
     }
 
     /// Returns the next event from a special queue
+    #[deprecated(note = "Broken API: use `poll_for_special_event2` instead")]
     #[cfg(any(feature = "xinput", feature = "present"))]
+    #[allow(deprecated)]
     pub fn poll_for_special_event(&self, se: SpecialEventId) -> Result<Option<Event>> {
         unsafe {
             let ev = xcb_poll_for_special_event(self.c, se.raw);
@@ -1323,6 +1355,59 @@ impl Connection {
         }
     }
 
+    /// Start listening for a special event.
+    ///
+    /// Effectively creates an internal special queue for this event
+    /// XGE events are only defined in the `xinput` and `present` extensions
+    ///
+    /// This function is present only if either of the `xinput` or `present` cargo features are active.
+    #[cfg(any(feature = "xinput", feature = "present"))]
+    pub fn register_for_special_event<EID: Xid>(
+        &self,
+        extension: Extension,
+        eid: EID,
+    ) -> SpecialEvent {
+        unsafe {
+            let ext: *mut xcb_extension_t = match extension {
+                #[cfg(feature = "xinput")]
+                Extension::Input => &mut xinput::FFI_EXT as *mut _,
+                #[cfg(feature = "present")]
+                Extension::Present => &mut present::FFI_EXT as *mut _,
+                _ => panic!("only Input and Present have XGE events"),
+            };
+
+            let raw = xcb_register_for_special_xge(self.c, ext, eid.resource_id(), ptr::null_mut());
+
+            SpecialEvent { raw }
+        }
+    }
+
+    /// Stop listening to a special event
+    #[cfg(any(feature = "xinput", feature = "present"))]
+    pub fn unregister_for_special_event(&self, se: SpecialEvent) {
+        unsafe {
+            xcb_unregister_for_special_event(self.c, se.raw);
+        }
+    }
+
+    /// Returns the next event from a special queue, blocking until one arrives
+    #[cfg(any(feature = "xinput", feature = "present"))]
+    pub fn wait_for_special_event2(&self, se: &SpecialEvent) -> Result<Event> {
+        unsafe {
+            let ev = xcb_wait_for_special_event(self.c, se.raw);
+            self.handle_wait_for_event(ev)
+        }
+    }
+
+    /// Returns the next event from a special queue
+    #[cfg(any(feature = "xinput", feature = "present"))]
+    pub fn poll_for_special_event2(&self, se: &SpecialEvent) -> Result<Option<Event>> {
+        unsafe {
+            let ev = xcb_poll_for_special_event(self.c, se.raw);
+            self.handle_poll_for_event(ev)
+        }
+    }
+
     /// Discards the reply for a request.
     ///
     /// Discards the reply for a request. Additionally, any error generated
@@ -1533,7 +1618,10 @@ impl Connection {
         self.check_request(self.send_request_checked(req))
     }
 
-    /// Get the reply of a previous request, or an error if one occured.
+    /// Gets the reply of a previous request, or an error if one occurred.
+    ///
+    /// This is blocking; it does not return until the reply has been received. For the non-blocking
+    /// version, see [`poll_for_reply`].
     ///
     /// # Example
     /// ```no_run
@@ -1550,31 +1638,25 @@ impl Connection {
     /// #   Ok(())
     /// # }
     /// ```
+    ///
+    /// [`poll_for_reply`]: Self::poll_for_reply
     pub fn wait_for_reply<C>(&self, cookie: C) -> Result<C::Reply>
     where
         C: CookieWithReplyChecked,
     {
         unsafe {
-            let mut error: *mut xcb_generic_error_t = std::ptr::null_mut();
+            let mut error: *mut xcb_generic_error_t = ptr::null_mut();
             let reply = xcb_wait_for_reply64(self.c, cookie.sequence(), &mut error as *mut _);
-            match (reply.is_null(), error.is_null()) {
-                (true, true) => {
-                    self.has_error()?;
-                    unreachable!("xcb_wait_for_reply64 returned null without I/O error");
-                }
-                (true, false) => {
-                    let error = error::resolve_error(error, &self.ext_data);
-                    Err(error.into())
-                }
-                (false, true) => Ok(C::Reply::from_raw(reply as *const u8)),
-                (false, false) => unreachable!("xcb_wait_for_reply64 returned two pointers"),
-            }
+            self.handle_reply_checked::<C>(reply, error)
         }
     }
 
     /// Get the reply of a previous unchecked request.
     ///
-    /// If an error occured, `None` is returned and the error will be delivered to the event loop.
+    /// If an error occurred, `None` is returned and the error will be delivered to the event loop.
+    ///
+    /// This is blocking; it does not return until the reply has been received. For the non-blocking
+    /// version, see [`poll_for_reply_unchecked`].
     ///
     /// # Example
     /// ```no_run
@@ -1591,17 +1673,184 @@ impl Connection {
     /// #   Ok(())
     /// # }
     /// ```
+    ///
+    /// [`poll_for_reply_unchecked`]: Self::poll_for_reply_unchecked
     pub fn wait_for_reply_unchecked<C>(&self, cookie: C) -> ConnResult<Option<C::Reply>>
     where
         C: CookieWithReplyUnchecked,
     {
         unsafe {
             let reply = xcb_wait_for_reply64(self.c, cookie.sequence(), ptr::null_mut());
-            if reply.is_null() {
-                self.has_error()?;
-                Ok(None)
-            } else {
-                Ok(Some(C::Reply::from_raw(reply as *const u8)))
+            self.handle_reply_unchecked::<C>(reply)
+        }
+    }
+
+    /// Gets the reply of a previous request if it has been received, or an error if one occurred.
+    ///
+    /// This is non-blocking; if no reply has been received yet, it returns [`None`]. For the
+    /// blocking version, see [`wait_for_reply`].
+    ///
+    /// # Examples
+    /// ```no_run
+    /// # use xcb::x;
+    /// # fn main() -> xcb::Result<()> {
+    /// #     let (conn, screen_num) = xcb::Connection::connect(None)?;
+    /// let (wm_protocols_cookie, wm_name_cookie) = (
+    ///     conn.send_request(&x::InternAtom {
+    ///         only_if_exists: true,
+    ///         name: b"WM_PROTOCOLS",
+    ///     }),
+    ///     conn.send_request(&x::InternAtom {
+    ///         only_if_exists: true,
+    ///         name: b"WM_NAME",
+    ///     }),
+    /// );
+    /// let (wm_protocols_atom, wm_name_atom) = {
+    ///     let (
+    ///         mut wm_protocols_atom,
+    ///         mut wm_name_atom,
+    ///     ) = (None, None);
+    ///
+    ///     loop {
+    ///         // If `wm_protocols_atom` is yet to be received, poll for it.
+    ///         if wm_protocols_atom.is_none() {
+    ///             wm_protocols_atom = conn
+    ///                 .poll_for_reply(&wm_protocols_cookie)
+    ///                 .transpose()?
+    ///                 .map(|reply| reply.atom());
+    ///         }
+    ///         // If `wm_name_atom` is yet to be received, poll for it.
+    ///         if wm_name_atom.is_none() {
+    ///             wm_name_atom = conn
+    ///                 .poll_for_reply(&wm_name_cookie)
+    ///                 .transpose()?
+    ///                 .map(|reply| reply.atom());
+    ///         }
+    ///
+    ///         // If both `wm_protocols_atom` and `wm_name_atom` have been
+    ///         // received, break from the loop.
+    ///         if let (
+    ///             Some(wm_protocols_atom),
+    ///             Some(wm_name_atom),
+    ///         ) = (wm_protocols_atom, wm_name_atom) {
+    ///             break (wm_protocols_atom, wm_name_atom);
+    ///         }
+    ///     }
+    /// };
+    /// #     Ok(())
+    /// # }
+    /// ```
+    ///
+    /// [`wait_for_reply`]: Self::wait_for_reply
+    pub fn poll_for_reply<C>(&self, cookie: &C) -> Option<Result<C::Reply>>
+    where
+        C: CookieWithReplyChecked,
+    {
+        unsafe {
+            let mut error: *mut xcb_generic_error_t = ptr::null_mut();
+            let mut reply: *mut c_void = ptr::null_mut();
+
+            let received = xcb_poll_for_reply64(
+                self.c,
+                cookie.sequence(),
+                &mut reply as *mut _,
+                &mut error as *mut _,
+            );
+
+            match received {
+                0 => None,
+                1 => Some(self.handle_reply_checked::<C>(reply, error)),
+                _ => panic!("unexpected return value from xcb_poll_for_reply64"),
+            }
+        }
+    }
+
+    /// Gets the reply of a previous unchecked request if it has been received.
+    ///
+    /// If an error occurred, [`None`] is returned and the error is delivered to the event loop.
+    ///
+    /// This is non-blocking; if no reply has been received yet, it returns
+    /// <code>[Some]\([None])</code>. For the blocking version, see [`wait_for_reply_unchecked`].
+    ///
+    /// # Examples
+    /// ```no_run
+    /// # use xcb::x;
+    /// # fn main() -> xcb::Result<()> {
+    /// #     let (conn, screen_num) = xcb::Connection::connect(None)?;
+    /// let (wm_protocols_cookie, wm_name_cookie) = (
+    ///     conn.send_request_unchecked(&x::InternAtom {
+    ///         only_if_exists: true,
+    ///         name: b"WM_PROTOCOLS",
+    ///     }),
+    ///     conn.send_request_unchecked(&x::InternAtom {
+    ///         only_if_exists: true,
+    ///         name: b"WM_NAME",
+    ///     }),
+    /// );
+    /// let (wm_protocols_atom, wm_name_atom) = {
+    ///     let (
+    ///         mut wm_protocols_atom,
+    ///         mut wm_name_atom,
+    ///     ) = (Some(None), Some(None));
+    ///
+    ///     loop {
+    ///         // If `wm_protocols_atom` is yet to be received, poll for it.
+    ///         if let Some(None) = wm_protocols_atom {
+    ///             wm_protocols_atom = conn
+    ///                 // connection error may happen
+    ///                 .poll_for_reply_unchecked(&wm_protocols_cookie)
+    ///                 .transpose()?
+    ///                 .map(|result| result.map(|reply| reply.atom()));
+    ///         }
+    ///         // If `wm_name_atom` is yet to be received, poll for it.
+    ///         if let Some(None) = wm_name_atom {
+    ///             wm_name_atom = conn
+    ///                 // connection error may happen
+    ///                 .poll_for_reply_unchecked(&wm_name_cookie)
+    ///                 .transpose()?
+    ///                 .map(|result| result.map(|reply| reply.atom()));
+    ///         }
+    ///
+    ///         match (wm_protocols_atom, wm_name_atom) {
+    ///             // If either `wm_protocols_atom` or `wm_name_atom` hasn't
+    ///             // been received, continue the loop.
+    ///             (Some(None), _) | (_, Some(None)) => continue,
+    ///
+    ///             // Otherwise, if both have been received, break from the
+    ///             // loop.
+    ///             (
+    ///                 wm_protocols_atom,
+    ///                 wm_name_atom,
+    ///             ) => break (
+    ///                 wm_protocols_atom.flatten(),
+    ///                 wm_name_atom.flatten(),
+    ///             ),
+    ///         }
+    ///     }
+    /// };
+    /// #     Ok(())
+    /// # }
+    /// ```
+    ///
+    /// [`wait_for_reply_unchecked`]: Self::wait_for_reply_unchecked
+    pub fn poll_for_reply_unchecked<C>(&self, cookie: &C) -> Option<ConnResult<Option<C::Reply>>>
+    where
+        C: CookieWithReplyUnchecked,
+    {
+        unsafe {
+            let mut reply: *mut c_void = ptr::null_mut();
+
+            let received = xcb_poll_for_reply64(
+                self.c,
+                cookie.sequence(),
+                &mut reply as *mut _,
+                ptr::null_mut(),
+            );
+
+            match received {
+                0 => None,
+                1 => Some(self.handle_reply_unchecked::<C>(reply)),
+                _ => panic!("unexpected return value from xcb_poll_for_reply64"),
             }
         }
     }
@@ -1612,6 +1861,9 @@ impl Connection {
     ///
     /// This retrieves the total number of bytes read from this connection,
     /// to be used for diagnostic/monitoring/informative purposes.
+    ///
+    /// Since: libxcb 1.14
+    #[cfg(feature = "libxcb_v1_14")]
     pub fn total_read(&self) -> usize {
         unsafe { xcb_total_read(self.c) as usize }
     }
@@ -1622,11 +1874,72 @@ impl Connection {
     ///
     /// This retrieves the total number of bytes written to this connection,
     /// to be used for diagnostic/monitoring/informative purposes.
+    ///
+    /// Since: libxcb 1.14
+    #[cfg(feature = "libxcb_v1_14")]
     pub fn total_written(&self) -> usize {
         unsafe { xcb_total_written(self.c) as usize }
     }
 }
 
+impl Connection {
+    unsafe fn handle_wait_for_event(&self, ev: *mut xcb_generic_event_t) -> Result<Event> {
+        if ev.is_null() {
+            self.has_error()?;
+            panic!("xcb_wait_for_event returned null with I/O error");
+        } else if is_error(ev) {
+            Err(error::resolve_error(ev as *mut _, &self.ext_data).into())
+        } else {
+            Ok(event::resolve_event(ev, &self.ext_data))
+        }
+    }
+
+    unsafe fn handle_poll_for_event(&self, ev: *mut xcb_generic_event_t) -> Result<Option<Event>> {
+        if ev.is_null() {
+            self.has_error()?;
+            Ok(None)
+        } else if is_error(ev) {
+            Err(error::resolve_error(ev as *mut _, &self.ext_data).into())
+        } else {
+            Ok(Some(event::resolve_event(ev, &self.ext_data)))
+        }
+    }
+
+    unsafe fn handle_reply_checked<C>(
+        &self,
+        reply: *mut c_void,
+        error: *mut xcb_generic_error_t,
+    ) -> Result<C::Reply>
+    where
+        C: CookieWithReplyChecked,
+    {
+        match (reply.is_null(), error.is_null()) {
+            (true, true) => {
+                self.has_error()?;
+                unreachable!("xcb_wait_for_reply64 returned null without I/O error");
+            }
+            (true, false) => {
+                let error = error::resolve_error(error, &self.ext_data);
+                Err(error.into())
+            }
+            (false, true) => Ok(C::Reply::from_raw(reply as *const u8)),
+            (false, false) => unreachable!("xcb_wait_for_reply64 returned two pointers"),
+        }
+    }
+
+    unsafe fn handle_reply_unchecked<C>(&self, reply: *mut c_void) -> ConnResult<Option<C::Reply>>
+    where
+        C: CookieWithReplyUnchecked,
+    {
+        if reply.is_null() {
+            self.has_error()?;
+            Ok(None)
+        } else {
+            Ok(Some(C::Reply::from_raw(reply as *const u8)))
+        }
+    }
+}
+
 impl AsRef<Connection> for Connection {
     fn as_ref(&self) -> &Connection {
         self
@@ -1750,6 +2063,17 @@ pub(crate) fn align_pad(base: usize, align: usize) -> usize {
     (-base & (align - 1)) as usize
 }
 
+/// Get a value from a pointer, using ptr::copy_nonoverlapping to avoid alignment issues
+///
+/// # Safety
+/// The pointer must point to a valid `T` value
+#[inline]
+pub(crate) unsafe fn value_from_ptr<T>(ptr: *const T) -> T {
+    let mut val = mem::MaybeUninit::<T>::uninit();
+    ptr::copy_nonoverlapping(ptr, val.as_mut_ptr(), 1);
+    val.assume_init()
+}
+
 #[test]
 fn test_align_pad() {
     // align 1
